#!/usr/bin/perl -w

# BEGIN LICENSE BLOCK
#
# Copyright (c) 1996-2003 Jesse Vincent <jesse@bestpractical.com>
#
# (Except where explictly superceded by other copyright notices)
#
# This work is made available to you under the terms of Version 2 of
# the GNU General Public License. A copy of that license should have
# been provided with this software, but in any event can be snarfed
# from www.gnu.org.
#
# This work is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# Unless otherwise specified, all modifications, corrections or
# extensions to this work which alter its source code become the
# property of Best Practical Solutions, LLC when submitted for
# inclusion in the work.
#
#
# END LICENSE BLOCK


use strict;

use vars qw/$VERSION/;

$VERSION = '0.1';

$SOAP::Constants::DO_NOT_USE_XML_PARSER = 1;

use lib ( "/opt/rt3/lib", "/opt/rt3/local/lib" );
use RT::EmailParser;

sub RT::Date::AsISO8601 {
    my $self = shift;
    my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday ) =
      gmtime( $self->Unix );
    if ( $self->Unix > 0 ) {
        return sprintf( "%04d-%02d-%02dT%02d:%02d:%02dZ",
                        ( $year + 1900 ),
                        ( $mon + 1 ),
                        $mday, $hour, $min, $sec );
    }
    else {
        return '';
    }
}

# {{{ RT object information for soap serialization

package RT::User;
our $SOAP_ATTRS = (
              { Name         => { method => 'Name',         type => 'string' },
                EmailAddress => { method => 'EmailAddress', type => 'string' },
                RealName     => { method => 'RealName',     type => 'string' },
                id           => { method => 'id',           type => 'int' }, }
);

sub SOAPAttrs { return $RT::User::SOAP_ATTRS; }

package RT::Transaction;

our $SOAP_ATTRS = (
    {  id          => { method => 'id',          type => 'int' },
       Description => { method => 'Description', type => 'string' },
       OldValue    => { method => 'OldValue',    type => 'string' },
       NewValue    => { method => 'NewValue',    type => 'string' },
       Attachments => { method => 'Attachments', type => 'array' },

    } );

sub SOAPAttrs { return $RT::Transaction::SOAP_ATTRS; }

package RT::Link;

our $SOAP_ATTRS = ( { id     => { method => 'id',     type => 'int' },
                      Base   => { method => 'Base',   type => 'string' },
                      Target => { method => 'Target', type => 'string' },
                      Type   => { method => 'Target', type => 'string' }, } );
sub SOAPAttrs { return $RT::Link::SOAP_ATTRS; }

package RT::Ticket;

$RT::Ticket::SOAP_ATTRS = (
      { id      => { method => 'id',                       type => 'int' },
        Subject => { method => 'Subject',                  type => 'string' },
        Status  => { method => 'Status',                   type => 'string' },
        Owner   => { method => 'OwnerObj' },
        Cc      => { method => 'Cc->UserMembersObj' },
        AdminCc => { method => 'AdminCc->UserMembersObj' },
        Requestors   => { method => 'Requestors->UserMembersObj' },
        DependsOn    => { method => 'DependsOn' },
        DependedOnBy => { method => 'DependedOnBy' },
        MemberOf     => { method => 'MemberOf' },
        Members      => { method => 'Members' },
        RefersTo     => { method => 'RefersTo' },
        ReferredToBy => { method => 'ReferredToBy' },
        Transactions => { method => 'Transactions' },
        Starts => { method => 'StartsObj->AsISO8601', type => 'datetime' },
        Started  => { method => 'StartedObj->AsISO8601',  type => 'datetime' },
        Due      => { method => 'DueObj->AsISO8601',      type => 'datetime' },
        Resolved => { method => 'ResolvedObj->AsISO8601', type => 'datetime' },
        Created  => { method => 'CreatedObj->AsISO8601',  type => 'datetime' },
        Told     => { method => 'ToldObj->AsISO8601',     type => 'datetime' },
        LastUpdated =>
          { method => 'LastUpdatedObj->AsISO8601', type => 'datetime' },
        Creator       => { method => 'CreatorObj->Name' },
        LastUpdatedBy => { method => 'LastUpdatedByObj->Name' },
        Queue         => { method => 'QueueObj->Name' },
        Type          => { method => 'Type' }, } );

sub SOAPAttrs { return $RT::Ticket::SOAP_ATTRS; }
# }}}

# {{{ Glue to turn on serializers 

*SOAP::Serializer::as_RT__Transaction = \&SOAP::Serializer::as_RT__Record;
*SOAP::Serializer::as_RT__User        = \&SOAP::Serializer::as_RT__Record;
*SOAP::Serializer::as_RT__Ticket      = \&SOAP::Serializer::as_RT__Record;
*SOAP::Serializer::as_RT__Link        = \&SOAP::Serializer::as_RT__Record;
*SOAP::Serializer::as_RT__Links = \&SOAP::Serializer::as_RT__SearchBuilder;
*SOAP::Serializer::as_RT__Transactions =
  \&SOAP::Serializer::as_RT__SearchBuilder;
*SOAP::Serializer::as_RT__Users   = \&SOAP::Serializer::as_RT__SearchBuilder;
*SOAP::Serializer::as_RT__Tickets = \&SOAP::Serializer::as_RT__SearchBuilder;
*SOAP::Serializer::as_RT__Attachments =
  \&SOAP::Serializer::as_RT__SearchBuilder;

# }}}

# {{{ Custom serializers 
sub SOAP::Serializer::as_RT__SearchBuilder {
    my ( $self, $object, $name, $type, $attrib ) = @_;

    $self->{'skip_attrs'} = $object->{'SOAP_skip_attrs'}
      unless ( $self->{'skip_attrs'} );

    $attrib->{'xmlns:bp'} = "http://schema.bestpractical.com/";
    $attrib->{'xsi:type'} = "bp:$type";

    my @serialized_content;
    while ( my $record = $object->Next ) {
        push @serialized_content, $self->encode_object($record);
    }
    my $retvals = [ $name, $attrib, [@serialized_content] ];

    return ($retvals);
}

sub SOAP::Serializer::as_RT__Record {
    my ( $self, $object, $name, $type, $attrib ) = @_;

    $self->{'skip_attrs'} = $object->{'SOAP_skip_attrs'}
      unless ( $self->{'skip_attrs'} );
    $attrib->{'xmlns:bp'} = "http://schema.bestpractical.com/";
    $attrib->{'xsi:type'} = "bp:$type";

    my @serialized_content;
    my $ATTRS = $object->SOAPAttrs;
    foreach my $attr ( keys %$ATTRS ) {
        next if ( $self->{'skip_attrs'}->{ ref($object) }->{$attr} );

        #print STDERR "Looking at ".ref($object).".".$attr."\n";
        my $local_object = $object;  #Since we may futz with it inside the loop;
             # though we do want it to reset on each iteration
        my $method = $ATTRS->{$attr}->{'method'};
        while ( $method =~ /^(.*?)\-\>(.*)$/ ) {
            $local_object = $local_object->$1() if ( $local_object->can($1) );
            $method = $2;
        }

        unless ( $local_object->can($method) ) {

#print STDERR "Can $local_object $method? -- ".($local_object->can($method)? 'yes': 'no'). "\n";
        }
        my $data;
        my $type    = $ATTRS->{$attr}->{'type'};
        my $content = $local_object->$method();

#We may want to serialize this object as something else to get it to use a different serializer
# SOAP lite should have some other facility for that.
        if (  
              UNIVERSAL::isa( $content, 'RT::SearchBuilder' ) ) {
            my @values = @{$content->ItemsArrayRef};
            $content = \@values;
        }

        $data =
          $self->encode_object( $content, $attr, $ATTRS->{$attr}->{'type'} );
        push @serialized_content, $data;
    }
    my $retvals = [ $name, $attrib, [@serialized_content] ];

    return ($retvals);
}
# }}}

package RT;

use RT;
use RT::Tickets;

# Load the config file
RT::LoadConfig();

#Connect to the database and get RT::SystemUser and RT::Nobody loaded
RT::Init();

package main;

my $port = pop (@ARGV) || 9000;
my $host = shift (@ARGV) || 'localhost';

use SOAP::Transport::HTTP;
use Hook::LexWrap;

Hook::LexWrap::wrap SOAP::Transport::HTTP::Server::handle, pre => sub {
    my $self = shift->new;
    my ($user,$pass);
    my $credentials = $self->request->headers->authorization_basic;
    if ($credentials =~ /^(.*?):(.*)$/) {
        $user = $1;
        $pass = $2;
    }
    $NETRT::CurrentUser  = RT::CurrentUser->new($user);

    unless ($NETRT::CurrentUser->id) {
        undef ($NETRT::CurrentUser);
        # this is how hook lexwrap lets us short circuit running the original sub
        $_[-1] = $self->response(HTTP::Response->new(401)) # FORBIDDEN
    }
    unless ($NETRT::CurrentUser->IsPassword($pass)) {
        undef ($NETRT::CurrentUser);
        $_[-1] = $self->response(HTTP::Response->new(401)) # FORBIDDEN

    }

    unshift @_, ref($self);
 };



print "Ready\n";
SOAP::Transport::HTTP::Daemon->new( LocalPort => $port, Reuse => 1 )
  ->dispatch_to('NETRT')->handle;

package NETRT;

use vars qw/$CurrentUser/;

our $DEBUG = 1;


# {{{ createTicket
sub createTicket {
    my $self = shift;
    my %args  = (
                @_);

    my @Actions;

    my $due = new RT::Date( $CurrentUser );
    $due->Set( Format => 'unknown', Value => $args{'Due'} );
    my $starts = new RT::Date( $CurrentUser);
    $starts->Set( Format => 'unknown', Value => $args{'Starts'} );


    my $ticket = RT::Ticket->new($CurrentUser);

    my $parser = RT::EmailParser->new();
    $parser->ParseMIMEEntityFromScalar($args{'MimeMessage'});

    my %create_args = (
        Queue           => $args{'Queue'},
        Owner           => $args{'Owner'},
        InitialPriority => $args{'InitialPriority'},
        FinalPriority   => $args{'FinalPriority'},
        TimeLeft        => $args{'TimeLeft'},
        TimeEstimated        => $args{'TimeEstimated'},
        TimeWorked      => $args{'TimeWorked'},
        Requestor       => $args{'Requestors'},
        Cc              => $args{'Cc'},
        AdminCc         => $args{'AdminCc'},
        Subject         => $args{'Subject'},
        Status          => ($args{'Status'} || 'new'),
        Due             => $due->ISO,
        Starts          => $starts->ISO,
        MIMEObj         => $parser->Entity
    );
  foreach my $arg (%args) {
        if ($arg =~ /^CustomField-(\d+)(.*?)$/) {
            $create_args{"CustomField-".$1} = $args{"$arg"};
        }
    }
    my ( $id, $Trans, $ErrMsg ) = $ticket->Create(%create_args);
    unless ( $id && $Trans ) {
        return({ status => $id, message => $ErrMsg});
    }
    my @linktypes = qw( DependsOn MemberOf RefersTo );

    foreach my $linktype (@linktypes) {
        foreach my $luri ( split ( / /, $args{"new-$linktype"} ) ) {
            $luri =~ s/\s*$//;    # Strip trailing whitespace
            my ( $val, $msg ) = $ticket->AddLink( Target => $luri, Type   => $linktype);
            push ( @Actions,  { status => $val, message => $msg }) unless ($val);
        }

        foreach my $luri ( split ( / /, $args{$linktype."-new"} ) ) {
            $luri =~ s/\s*$//;    # Strip trailing whitespace
            my ( $val, $msg ) = $ticket->AddLink( Base => $luri, Type => $linktype);
            push ( @Actions,  { status => $val, message => $msg }) unless ($val);
       }
   }


            return ( { status => $id, message => $ErrMsg });
}
# }}}

# {{{ getTicket
sub getTicket {
    my $self = shift;
    my $id   = shift;

    print "in getTicket for $id\n";
    my $ticket = RT::Ticket->new($CurrentUser);
    $ticket->Load($id);
    return ($ticket);
}

# }}}

# {{{ getTickets 
sub getTickets {
    my $self  = shift;
    my $query = shift;
    print "in getTickets for $query\n";
    local $RT::Ticket::SOAP_ATTRS = $RT::Ticket::SOAP_ATTRS;
    delete $RT::Ticket::SOAP_ATTRS->{'Transactions'};
    my $tix = RT::Tickets->new($CurrentUser);
    $tix->FromSQL($query);

    foreach my $ticket (@{$tix->ItemsArrayRef}) {
        print "Looking at ticket ".$ticket->id ."\n";
    }
    return (@{$tix->ItemsArrayRef});
}
# }}}

# {{{ updateTicket
sub updateTicket {
    my $self = shift;
    my %args = (@_);

    my $ticket = RT::Ticket->new($CurrentUser);
    $ticket->Load( $args{'id'} );
    unless ( $ticket->id ) {
        return ( { status => 0, message => "Invalid ticket id" } );
    }

    my $parser = RT::EmailParser->new();
    $parser->ParseMIMEEntityFromScalar($args{'MimeMessage'});

    if ( $args{'UpdateType'} eq 'comment' ) {
        my ( $val, $msg ) = $ticket->Comment( MIMEObj => $parser->Entity,
                                              TimeWorked => $args{'TimeWorked'},
                                              CcMessageTo  => $args{'Cc'},
                                              BccMessageTo => $args{'Bcc'} );

        return ( { status => $val, message => $msg } ) unless ($val);

    }
    elsif ( $args{'UpdateType'} eq 'reply' ) {
        my ( $val, $msg ) = $ticket->Correspond( MIMEObj => $parser->Entity,
                                              TimeWorked => $args{'TimeWorked'},
                                              CcMessageTo  => $args{'Cc'},
                                              BccMessageTo => $args{'Bcc'} );

        return ( { status => $val, message => $msg } ) unless ($val);
    }
    else {
        return ( { status => 0, message => 'Update type unknown' } );
    }

    if ( $args{'Status'} && $args{'Status'} ne $ticket->Status ) {
        my ( $val, $msg ) = $ticket->SetStatus( $args{'Status'} );

        return ( { status => $val, message => $msg } ) unless ($val);
    }

    return ( { status => 1, message => 'Ticket updated' } );
}
# }}}

# {{{ modifyTicket

sub modifyTicket {
        my $self = shift;
    my %args = (@_);
  # {{{ Set basic fields 
    my @attribs = qw(
      Subject
      FinalPriority
      Priority
      TimeEstimated
      TimeWorked
      TimeLeft
      Status
      Queue
    );



    my $ticket = RT::Ticket->new($CurrentUser);
    $ticket->Load( $args{'id'} );
    unless ( $ticket->id ) {
      return ( { status => 0, message => "Invalid ticket id" } );
    }


    my @return_values;

    if ( $args{'Queue'} and ( $args{'Queue'} !~ /^(\d+)$/ ) ) {
        my $tempqueue = RT::Queue->new($CurrentUser);
        $tempqueue->Load( $args{'Queue'} );
        if ( $tempqueue->id ) {
            $args{'Queue'} = $tempqueue->Id();
        }
    }

    @return_values = RT::SOAP::Private::_UpdateRecordObject(
        AttributesRef => \@attribs,
        Object        => $ticket,
        ARGSRef       => \%args,
    );

    # We special case owner changing, so we can use ForceOwnerChange
    if ( $args{'Owner'} && ( $ticket->Owner != $args{'Owner'} ) ) {
        my ($ChownType);
        if ( $args{'ForceOwnerChange'} ) {
            $ChownType = "Force";
        }
        else {
            $ChownType = "Give";
        }

        my ( $val, $msg ) =
          $ticket->SetOwner( $args{'Owner'}, $ChownType );
        push ( @return_values, { attrib => 'Owner', message => $msg, status => $val} );
    }
    # }}}

    
    # {{{ Set custom fields
    my $custom_fields_to_mod;
    foreach my $arg ( keys %args ) {
        if ( $arg =~ /^CustomField-(\d+)-/ ) {

            # For each of those tickets, find out what custom fields we want to work with.
            $custom_fields_to_mod->{$1} = 1;
        }
    }

    # For each custom field
    foreach my $cf ( keys %{$custom_fields_to_mod} ) {

    foreach my $arg ( keys %args) {
        next unless ( $arg =~ /^CustomField-$cf-/ );
        my @values = ( ref( $args{$arg} ) eq 'ARRAY' ) ? @{ $args{$arg} } : ( $args{$arg} );
        if ( ( $arg =~ /-AddValue$/ ) || ( $arg =~ /-Value$/ ) ) {
            foreach my $value (@values) {
                next unless ($value);
                my ( $val, $msg ) = $ticket->AddCustomFieldValue(Field => $cf,
                                                                 Value => $value
                );
                push ( @return_values, $msg );
            }
        }
        elsif ( $arg =~ /-DeleteValues$/ ) {
            foreach my $value (@values) {
                next unless ($value);
                my ( $val, $msg ) = $ticket->DeleteCustomFieldValue(
                                                                 Field => $cf,
                                                                 Value => $value
                );
                push ( @return_values, $msg );
            }
        }
        elsif ( $arg =~ /-Values/ ) {
            my $cf_values = $ticket->CustomFieldValues($cf);

            my %values_hash;
            foreach my $value (@values) {
                next unless ($value);

                # build up a hash of values that the new set has
                $values_hash{$value} = 1;

                unless ( $cf_values->HasEntry($value) ) {
                    my ( $val, $msg ) = $ticket->AddCustomFieldValue(
                                                                 Field => $cf,
                                                                 Value => $value
                    );
                    push ( @return_values, $msg );
                }

            }
            while ( my $cf_value = $cf_values->Next ) {
                unless ( $values_hash{ $cf_value->Content } == 1 ) {
                    my ( $val, $msg ) = $ticket->DeleteCustomFieldValue(
                                                     Field => $cf,
                                                     Value => $cf_value->Content
                    );
                    push ( @return_values, $msg );

                }

            }
        }
        else {
            push ( @return_values,
                   "User asked for an unknown update type for custom field "
                     . $cf->Name
                     . " for ticket "
                     . $ticket->id );
        }
    }
}


    # }}} end of custom fields stuff


# {{{ Deal with watchers
   foreach my $key ( keys %args ) {
    if ($key =~ /^(Requestors|Cc|AdminCc)$/) {
        my $type = $1;
        my $members = $ticket->$1()->UserMembersObj;
        my @current = map {$_->Name } @{$members->ItemsArrayRef};
        my @new = @{$args{$key}};


        my (@to_add, @to_delete) ;
        my %seen;
        foreach my $item (@current) {
            $seen{$item} = 1; 
        }
        foreach my $item(@new) {
            $seen{$item} += 2;
        }
        foreach my $item (keys %seen) {
            push @to_delete, $item if ($seen{$item} == 1) ;
            push @to_add, $item if ($seen{$item} == 2) ;
        }    
   
        # TODO Requestors changes to Requestor. this is lame. Fix in the core to accept both 
        $type =~ s/s$//;

        foreach my $item (@to_delete) {
            my ($id, $msg) = $ticket->DeleteWatcher(Type => $type, Email => $item);
            push (@return_values, { message => $msg, status => $id, attribute => "Delete".$type, value => $item});
        }

        foreach my $item (@to_add) {
            my ($id, $msg) = $ticket->AddWatcher(Type => $type, Email => $item);
            push (@return_values, { message => $msg, status => $id, attribute => "Add".$type, value => $item});
        }


    }
        

   } 

# }}}



    return @return_values;

}
# }}}


package RT::SOAP::Private;

# {{{ _UpdateRecordObj 
=head2 _UpdateRecordObj ( ARGSRef => \%ARGS, Object => RT::Record, AttributesRef =>
 \@attribs)

@attribs is a list of ticket fields to check and update if they differ from the  B
<Object>'s current values. ARGSRef is a ref to a paramhash.

Returns an array of success/failure hashes

=cut

sub _UpdateRecordObject {
    my %args = (
        ARGSRef       => undef,
        AttributesRef => undef,
        Object        => undef,
        AttributePrefix => undef,
        @_
    );

    my (@results);

    my $object     = $args{'Object'};
    my $attributes = $args{'AttributesRef'};
    my $ARGSRef    = $args{'ARGSRef'};
    foreach my $attribute (@$attributes) {
        my $value;
        if ( defined $ARGSRef->{$attribute} ) {
            $value = $ARGSRef->{$attribute};
        }
        elsif ( defined( $args{'AttributePrefix'} ) &&
                defined( $ARGSRef->{ $args{'AttributePrefix'} . "-" . $attribute })) {
            $value = $ARGSRef->{ $args{'AttributePrefix'} . "-" . $attribute };

        } else {
                next;
        }

            $value =~ s/\r\n/\n/gs;
        if ( $value ne $object->$attribute() ) {
            my $method = "Set$attribute";
            my ( $status, $msg ) = $object->$method($value);

            push @results,
              { attribute => $attribute,
                status    => $status,
                message   => $msg };
        }
    }
    return (@results);
}
# }}}


=head1 THE PLAN


Create wrapper objects for:

    Ticket
    User
    Group
    Queue
    Transaction
    Attachment

=cut

